import waba.fx.*;     // Graphics
//import waba.io.*;
import waba.sys.*;    // Convert
import waba.ui.*;
//import waba.util.*;
import extra.ui.*;  // wextras: List (Popup, RelativeContainer, PreferredSize), Title
// Random

/**
 * @author <A HREF="mailto:weyer@kagi.com">Steve Weyer</A>
 * Copyright (c) 2001, S. Weyer. All right reserved.
 * @version  waba 1.0
 * !!!a 08 Feb 2001
 * <A HREF="http://www.kagi.com/weyer/#Waba">Steve's other Waba apps</A>
 *
 * simplified/adapted/extended from a Java version:
 * <A HREF="http://www.videoactv.com/java/Mines/Mines.html">Videoactv Studios, L.C.</A> Copyright (c) 1996
 * <A HREF="mailto:dmoore@videoactv.com">original author</A>
 *
 * notes:
 * ! Newton: heap(vmstack): 6000, class size: 22000
 * ? better graphics
 * ? less recursive to avoid stack problems
 * ? inverted graphics on Palm
**/

public class Mines extends MainWindow {
  public static final boolean DEBUG = false;

  private static final int COMPHT = 15;
  private static final int COMPWID = 30; // 5 across
  private static final int COMPSEP = 2;
  private static final String[] NUM_MINES = {"10", "20", "30", "40", "50"};

  private Label timeLabel;
  private Button newButton;
  private List numList;
  private Label numLabel;
  private MineCanvas canvas;
  private Timer secTimer = null;

  int time = 0, mines = 10;

  public Mines() {
    // original version had a separate score panel -- just merge that here
    // top row: Title, time, New Game, # mines
    int objX = 0, objY = 0, objWid = COMPWID, objHt = COMPHT;
    Title title = new Title("Mines");
    title.setRect(objX, objY, this.width, COMPHT+1);
    add(title);

    objX += objWid + COMPSEP + 5; // shift 5, borrow from # at end of row
    newButton = new Button("New");
    newButton.setRect(objX, objY, objWid, objHt-2);
    add(newButton);

    objX += objWid + COMPSEP;
    timeLabel = new Label("", Label.CENTER);
    timeLabel.setRect(objX, objY, objWid, objHt);
    add(timeLabel);

    objX += objWid + COMPSEP;
    numList = new List(NUM_MINES);
    //numList.setSelectedIndex(0);
    numList.setRect(objX, objY+1, objWid, objHt-1);
    add(numList);

    objX += objWid + COMPSEP;
    numLabel = new Label("", Label.CENTER);
    numLabel.setRect(objX, objY, objWid-5, objHt);
    add(numLabel);

    objX =  0; objWid = this.width;
    objY = 16; objHt  = this.height-objY;
    canvas = new MineCanvas(objWid, objHt, this);
    canvas.setRect(objX, objY, objWid, objHt);
    add(canvas);

    newGame(mines);
  }

  public void newGame(int n) { // called at startup, New, <#mines>
    updateTime(0);
    updateMines(mines = n);
    canvas.newGame(mines);
  }

  public void updateTime(int newTime) {
    time = newTime;
    int mins = time / 60, secs = time % 60;
    timeLabel.setText(String.valueOf(mins) + ':' + ((secs < 10) ? "0" : "") + secs);
  }

  public void updateMines(int newMine) {
    numLabel.setText(String.valueOf(newMine));
  }

  public void startGame() { // called after 1st guess (click)
    if (secTimer == null)
      secTimer = addTimer(1000);
  }

  public void endGame(String msg) { // called after win/lose
    if (secTimer != null)
      removeTimer(secTimer);
    secTimer = null;
    Sound.beep();
    numLabel.setText(msg);
  }

  public void onEvent(Event ev) {
    Object obj = ev.target;
    int evType = ev.type;
    if (evType == ControlEvent.TIMER) // obj == secTimer
      updateTime(time+1);
    else if (evType == ControlEvent.PRESSED)
      if (obj == newButton)
        newGame(mines);
      else if (obj == numList)
        newGame(Convert.toInt(numList.getSelected()));
  }
}


class MineCanvas extends Control {
  // original version had separate Board class -- merge that
  final static int GRID_SIZE = 16;
  // overall game state
  final static int OK = 0;
  final static int WON = 1;
  final static int LOST = 2;

  Image upSquare, downSquare, bomb, xedBomb, flag, boom;
  Mines theApp;
  Graphics drawg = null;
  int cols, rows, num_mines;
  int gameState = LOST;
  int lastX, lastY;

  // board state
  static final byte UNTOUCHED = 3;
  static final byte MARKED = 4;
  static final byte UNCOVERED = 5;

  protected boolean[][] grid;
  public byte[][] boardState;
  public byte[][] neighbors;
  private Random rand;

  public MineCanvas(int wid, int ht, Mines app) {
    theApp = app;
    rand = new Random();

    cols = wid / GRID_SIZE;
    rows = ht / GRID_SIZE;

    upSquare = new Image("upsqr16.bmp");
    downSquare = new Image("insqr16.bmp");
    bomb =  new Image("bomb16.bmp");
    flag = new Image("flag16.bmp");
    xedBomb = new Image("xedbomb16.bmp");
    boom = new Image("boom16.bmp");
  }

  public void newGame(int n) {
    gameState = OK;
    grid = new boolean[cols][rows];
    boardState = new byte[cols][rows];
    neighbors = new byte[cols][rows];
    lastX = lastY = -1; // for first pen tap

    num_mines = n;
    int x, y;
    while (n > 0) {
      x = rand.getRandom(cols);
      y = rand.getRandom(rows);
      if (grid[x][y])
	continue;

      grid[x][y] = true;
      n--;
    };

if (Mines.DEBUG) System.out.print("\nBoard");
    int xmin, xmax, ymin = 0, ymax;
    for (y = 0; y < rows; y++) {
      ymax = ((y == rows-1) ? y : y+1);
      xmin = 0;
if (Mines.DEBUG) System.out.print("\nrow[" + y + "]: ");
      for (x = 0; x < cols; x++) {
        boardState[x][y] = UNTOUCHED;
if (Mines.DEBUG) System.out.print((grid[x][y]) ?"B," : " ,");
	if (grid[x][y]) {
          xmax = ((x == cols-1) ? x : x+1);
	  for (int i = xmin; i <= xmax; i++)
	    for (int j = ymin; j <= ymax; j++)
	      neighbors[i][j] += 1;
	};
        xmin = x;
      };
      ymin = y;
    };

if (Mines.DEBUG) {
      System.out.print("\nNeighbors");
      for (y=0; y < rows; y++) {
        System.out.print("\nrow[" + y + "]: ");
        for (x=0; x < cols; x++)
          System.out.print(String.valueOf(neighbors[x][y]) + ',');
      };
      System.out.println();
    };

    repaint();
  }


  public void onPaint(Graphics g) {
    //?? no need to erase since entire area filled initially -- but Palm is inverting!
    //g.setColor(255, 255, 255);
    //g.fillRect(0, 0, this.width, this.height);
    if (drawg == null) { // init here for later updates
      drawg = createGraphics();
      drawg.setClip(0, 0, this.width, this.height);
      //drawg.setDrawOp(Graphics.DRAW_OVER);
      //drawg.setColor(255, 255, 255);
      //drawg.fillRect(0, 0, this.width, this.height);
      //drawg.setColor(0, 0, 0);  // for text
    };

    for (int y = 0; y < rows; y++)
      for (int x = 0; x < cols; x++)
        squareChanged(x, y);
  }

  public void squareChanged (int x, int y) {    // mark, uncover, or just repainting
    int gx = x*GRID_SIZE, gy = y*GRID_SIZE, nbors;
    boolean isMine = grid[x][y]; //isMine(x,y);
    Image img = null;
    switch (boardState[x][y]) {
      case UNCOVERED:
        if (isMine) {
          img = (x == lastX && y == lastY) ? boom : bomb;
          break;
        }
        else {
          drawg.drawImage(downSquare, gx, gy);
          if ((nbors = neighbors[x][y]) != 0)
            drawg.drawText(String.valueOf(nbors), gx+4, gy+4);  // offset a little
          return;
        }
      case UNTOUCHED:
        img = (gameState == LOST && isMine) ? bomb : upSquare;
        break;
      case MARKED:
        img = (gameState == LOST && !isMine) ? xedBomb : flag;
        //break;
    };
    drawg.drawImage(img, gx, gy);
  }

  public void onEvent(Event ev) {
    if (gameState != OK || ! (ev instanceof PenEvent))
      return;
    int curX = ((PenEvent)ev).x / GRID_SIZE;
    int curY = ((PenEvent)ev).y / GRID_SIZE;
    // checking shouldn't be necessary unless non-std size
    if (curX < 0 || curX >= cols || curY < 0 || curY >= rows)
      return;

    if (ev.type == PenEvent.PEN_DOWN) {
      if (lastX < 0)
        theApp.startGame(); // very first click
      // save for PEN_UP
      lastX = curX;
      lastY = curY;
      return;
    };

    byte st = boardState[lastX][lastY];
    if (ev.type == PenEvent.PEN_UP) {
      if (curX == lastX && curY == lastY) {
        // same square. uncover it (if not already marked/uncovered)
        if (st == UNTOUCHED) {
          uncover(lastX, lastY);
          if (grid[lastX][lastY]) { //(isMine(lastX, lastY)) {
            gameState = LOST;
            theApp.endGame("Lost!");
            repaint();
            return;
          };
        };
      }
      else {
        // drag to diff square. (un)mark the orig square and decrement(inc) the count
        //st = mark(lastX,lastY);
        //if (st == MARKED)
        //  num_mines--;
	//else if (st == UNTOUCHED)
        //  num_mines++;

        switch (st) {
          case UNTOUCHED:
            st = MARKED;    // flag
            num_mines--;
            break;
          case MARKED:
            st = UNTOUCHED; // unflag
            num_mines++;
            break;
          default:
            return;
          };
        boardState[lastX][lastY] = st;
        squareChanged(lastX, lastY);
        theApp.updateMines(num_mines);
      };

      if (Won()) {
        gameState = WON;
        theApp.endGame("Won!");
      };
    };
  }

  public void uncover(int x, int y) {
    // only recurse on squares that haven't been touched before.
    //  LU | U | RU
    // ------------
    //  L  | - | R
    // ------------
    //  LD | D | RD
    if (boardState[x][y] == UNTOUCHED) {
      boardState[x][y] = UNCOVERED;
      squareChanged(x,y);
      if (neighbors[x][y] == 0) {
        int x1 = x+1, y1 = y+1;
        if (x > 0) {
          uncover(x-1,y); // left
          if (y > 0)
            uncover(x-1,y-1); // left-up
          if (y1 < rows)
            uncover(x-1,y1);  // left-down
        };
        if (x1 < cols) {
          uncover(x1,y); // right
          if (y > 0)
            uncover(x1,y-1);  // right-up
          if (y1 < rows)
            uncover(x1,y1);  // right-down
        };
        if (y > 0)
          uncover(x,y-1);  // up
        if (y1 < rows)
          uncover(x,y1);  // down
      };
    };
  }

  public boolean Won() {
    byte st;
    for (int x = 0; x < cols; x++)
      for (int y = 0; y < rows; y++)
        if ((st = boardState[x][y]) == UNTOUCHED)
	  return false;
	else if (st == MARKED && ! grid[x][y]) // isMine
	  return false;

    return true;
  }

/*
  public byte mark (int x, int y) {
    byte st;
    switch (st = boardState[x][y]) {
      case UNTOUCHED:
        st = MARKED;
        break;
      case MARKED:
        st = UNTOUCHED;
        break;
      default:
        return st;
    };
    boardState[x][y] = st;
    squareChanged(x, y);
    return st;
  }

  public boolean isMine(int x, int y) {
    //if (x < 0 || x >= cols || y < 0 || y >= rows)
    //  return false;
    return grid[x][y];
  }
*/
}


